(define (domain model-train-time-metric)
  
  (:requirements :typing :fluents)
  
  (:types train switch segment - object)
  
  (:predicates
   (idle ?t - train)
   ;; The train ?t is currently not moving.
   
   (head-segment ?t - train ?s - segment)
   ;; The head of the train ?t is in segment ?s.
   
   (tail-segment ?t - train ?s - segment)
   ;; The tail of the train ?t is in segment ?s.
   
   (visited-segment ?t - train ?s - segment)
   ;; Train ?t has ever "fully entered" segment ?s,
   ;; i.e, (tail-segment ?t ?s) was ever set.
   
   (switch-available ?sw - switch)
   ;; This switch can currently be used, i.e., there is no train
   ;; currently passing through it.
   
   (switch-entrance ?sw - switch ?s - segment)
   ;; Switch ?sw can currently be entered from segment ?s,
   ;; provided that it is available.
   
   (switch-exit ?sw - switch ?s - segment)
   ;; Switch ?sw can currently be left into segment ?s,
   ;; provided that it is available.
   
   (first-train-in-head-segment ?t - train)
   ;; Train ?t is the first train in its head segment, i.e, there
   ;; is no train (not even part of a train) in front of it in this
   ;; segment: the path to the next switch is clear.
   
   ;; Note: We can easily test if a train is the first train in its tail
   ;; segment by testing that either the head and tail segments are
   ;; different (in which case it extends over the switch its tail
   ;; is moving towards), or they are the same and it is the first train
   ;; in its head segment.
   
   (next-train ?pred - train ?succ - train)
   ;; Means that the tail segment of ?pred equals the head segment of ?succ
   ;; and there are no trains in between. In other words, this is set
   ;; when ?succ enters a segment that is not completely empty, and it is
   ;; cleared when ?pred leaves a segment that does not become completely
   ;; empty.
   
   (last-train-in-tail-segment ?t - train)
   ;; Train ?t is the last train in its tail segment, i.e., there is
   ;; nothing between the tail of train ?t and the switch that
   ;; is reached first when going from the tail of the train in the
   ;; opposite direction of the movement direction.
   
   ;; Note: We can easily test if a train is the last train in its head
   ;; segment by testing that either the head and tail segments are
   ;; different (in which case it extends over the switch its tail
   ;; is moving towards), or they are the same and it is the last train
   ;; in its tail segment.
   
   (segment-empty ?s - segment)
   ;; There is no train in segment ?s, not even part of a train.
   
   (SWITCH-HAS-ENTRANCE ?sw - switch ?s - segment)
   (SWITCH-HAS-EXIT ?sw - switch ?s - segment)
   )
  
  (:functions
   (head-segment-position ?t - train)
   (tail-segment-position ?t - train)
   ;; Note: These are both updated with (at start) effects for maximal
   ;; concurrency. This cannot lead to crashes (we hope) because
   ;; once we update the value, we can promise that we can move until
   ;; this distance, and noone can catch up with us because everyone
   ;; moves at the same speed.
   
   (SEGMENT-LENGTH ?s - segment)
   (TRAIN-LENGTH ?t - train)
   )
  

  ;; Advance a train which is the first train in its segment to the
  ;; front of its segment. This is only allowed if the train's tail
  ;; does not cross a segment boundary, otherwise we must advance its
  ;; tail to the next switch first.
  (:action advance-head-to-switch
    :parameters (?t - train ?s1 ?s2 - segment)
    :precondition (and (idle ?t)
                       (head-segment ?t ?s1)
                       (tail-segment ?t ?s2)
		    
		    ;; First train in head segment
                       (first-train-in-head-segment ?t)
		    
		    ;; Do not pull tail beyond segment boundary
		    ;; (special case s1 == s2 is covered unless (TRAIN-LENGTH ?t) == 0)
			(<= (- (SEGMENT-LENGTH ?s1)
			       (head-segment-position ?t))
			    (- (SEGMENT-LENGTH ?s2)
			       (tail-segment-position ?t)))
		    )
    :effect (and (not (idle ?t))
                 (idle ?t)
                 (assign (head-segment-position ?t) (SEGMENT-LENGTH ?s1))
                 (increase (tail-segment-position ?t)
                           (- (SEGMENT-LENGTH ?s1)
                              (head-segment-position ?t)))
                 )
    )
  
  ;; Advance a train which is not the first train in its head segment
  ;; as far as possible, i.e., to the tail of the preceding train in 
  ;; the same segment. This is only allowed if the train's tail
  ;; does not cross a segment boundary, otherwise we must advance its
  ;; tail to the next switch first.
  (:action advance-head-to-next-train
    :parameters (?t ?pred - train ?s1 ?s2 - segment)
    :precondition (and (idle ?t)
                       (head-segment ?t ?s1)
                       (tail-segment ?t ?s2)
		    
		    ;; Affected train is behind train ?pred
                       (next-train ?pred ?t)
                       (tail-segment ?pred ?s1) ;; should be implied
		    
		    ;; Do not pull tail beyond segment boundary
		    ;; (special case s1 == s2 is covered unless (TRAIN-LENGTH ?t) == 0)
			(<= (- (tail-segment-position ?pred)
			       (head-segment-position ?t))
			    (- (SEGMENT-LENGTH ?s2)
			       (tail-segment-position ?t)))
		    )
    :effect (and (not (idle ?t))
		 (idle ?t)
		 (assign (head-segment-position ?t)
                         (tail-segment-position ?pred))
		 (increase (tail-segment-position ?t)
                           (- (tail-segment-position ?pred)
			      (head-segment-position ?t)))
		 )
    )
  
  
  ;; Advance a train which is the first train in its head segment so that its
  ;; tail advances to the next switch. Dually to advance-head-to-switch,
  ;; this is only allowed if the head does not cross a segment
  ;; boundary, otherwise we must advance the head of the train first.
  (:action advance-tail-to-switch-for-leading-train
    :parameters (?t - train ?s1 ?s2 - segment)
    :precondition (and (idle ?t)
                       (head-segment ?t ?s1)
                       (tail-segment ?t ?s2)
		    
                       ;; Assumption: Train is first train in its head segment
                       (first-train-in-head-segment ?t)
                       
                       ;; Do not push head beyond segment boundary
                       (<= (- (SEGMENT-LENGTH ?s2)
                              (tail-segment-position ?t))
                           (- (SEGMENT-LENGTH ?s1)
                              (head-segment-position ?t)))
                       )
    :effect (and (not (idle ?t))
                 (idle ?t)
                 (increase (head-segment-position ?t)
                           (- (SEGMENT-LENGTH ?s2)
                              (tail-segment-position ?t)))
                 (assign (tail-segment-position ?t)
                         (SEGMENT-LENGTH ?s2))
                 )
    )
  
  ;; Advance a train which is NOT the first train in its head segment so that its
  ;; tail advances to the next switch. This is only allowed if the head does
  ;; not cross a segment boundary AND it does not crash into the tail of a
  ;; preceding train on the same segment.
  (:action advance-tail-to-switch-for-trailing-train
    :parameters (?t ?pred - train ?s1 ?s2 - segment)
    :precondition (and (idle ?t)
                       (head-segment ?t ?s1)
                       (tail-segment ?t ?s2)

                       ;; Assumption: Train is NOT first train in its head segment
                       (next-train ?pred ?t)
                       (tail-segment ?pred ?s1) ;; should be implied.
		    
                       ;; Do not push head into tail of preceding train
                       (<= (- (SEGMENT-LENGTH ?s2)
                              (tail-segment-position ?t))
                           (- (tail-segment-position ?pred)
                              (head-segment-position ?t)))
                       )
    :effect (and (not (idle ?t))
                 (idle ?t)
                 (increase (head-segment-position ?t)
                           (- (SEGMENT-LENGTH ?s2)
                              (tail-segment-position ?t)))
                 (assign (tail-segment-position ?t)
                         (SEGMENT-LENGTH ?s2))
                 )
    )


  ;; Advance head to the next segment (Case 1: New head segment is empty).
  (:action update-head-segment-to-empty-segment
    :parameters (?t - train ?s-old ?s-new - segment ?sw - switch)
    :precondition (and (idle ?t)
                       (head-segment ?t ?s-old)
                       (= (head-segment-position ?t)
                          (SEGMENT-LENGTH ?s-old))
                       (first-train-in-head-segment ?t) ;; should be implied
                       (segment-empty ?s-new)
                       (switch-entrance ?sw ?s-old)
                       (switch-exit ?sw ?s-new)
		    )
    :effect (and (not (idle ?t))
		 (idle ?t)
		 (not (head-segment ?t ?s-old))
		 (head-segment ?t ?s-new)
		 (assign (head-segment-position ?t) 0)
		 (not (segment-empty ?s-new))
		 (not (switch-available ?sw))
		 )	     
    )
  
  ;; Advance head to the next segment (Case 2: New head segment is NOT empty).
  (:action update-head-segment-to-occupied-segment
    :parameters (?t ?pred - train ?s-old ?s-new - segment ?sw - switch)
    :precondition (and (idle ?t)
                       (head-segment ?t ?s-old)
                       (= (head-segment-position ?t)
                          (SEGMENT-LENGTH ?s-old))
                       (first-train-in-head-segment ?t) ;; should be implied
                       (tail-segment ?pred ?s-new)
                       (last-train-in-tail-segment ?pred)
                       (switch-entrance ?sw ?s-old)
                       (switch-exit ?sw ?s-new)
                       )
    :effect (and (not (idle ?t))
		 (idle ?t)
		 (not (head-segment ?t ?s-old))
		 (head-segment ?t ?s-new)
		 (assign (head-segment-position ?t) 0)
		 (not (first-train-in-head-segment ?t))
		 (not (last-train-in-tail-segment ?pred))
		 (next-train ?pred ?t)
		 (not (switch-available ?sw))
		 )	     
    )


  ;; Advance tail to the next segment (Case 1: Old tail segment becomes empty).
  (:action update-tail-segment-leaving-empty-segment
    :parameters (?t - train ?s-old ?s-new - segment ?sw - switch)
    :precondition (and (idle ?t)
                       (tail-segment ?t ?s-old)
                       (= (tail-segment-position ?t)
                          (SEGMENT-LENGTH ?s-old))
                       (last-train-in-tail-segment ?t)
                       (switch-entrance ?sw ?s-old)
                       (switch-exit ?sw ?s-new)
                       )
    :effect (and (not (idle ?t))
		 (idle ?t)
		 (not (tail-segment ?t ?s-old))
		 (tail-segment ?t ?s-new)
		 (assign (tail-segment-position ?t) 0)
		 (last-train-in-tail-segment ?t) ;; redundant
		 (segment-empty ?s-old)
		 (switch-available ?sw)
		 (visited-segment ?t ?s-new)
		 )
    )
  
  ;; Advance tail to the next segment (Case 2: Old tail segment DOES NOT become empty)
  (:action update-tail-segment-leaving-occupied-segment
    :parameters (?t ?succ - train ?s-old ?s-new - segment ?sw - switch)
    :precondition (and (idle ?t)
                       (tail-segment ?t ?s-old)
                       (= (tail-segment-position ?t)
                          (SEGMENT-LENGTH ?s-old))
                       (head-segment ?succ ?s-old) ;; should be implied
                       (next-train ?t ?succ)
                       (switch-entrance ?sw ?s-old)
                       (switch-exit ?sw ?s-new)
                       )
    :effect (and (not (idle ?t))
		 (idle ?t)
		 (not (tail-segment ?t ?s-old))
		 (tail-segment ?t ?s-new)
		 (assign (tail-segment-position ?t) 0)
		 (last-train-in-tail-segment ?t)
		 (not (next-train ?t ?succ))
		 (first-train-in-head-segment ?succ)
		 (switch-available ?sw)
		 (visited-segment ?t ?s-new)
		 )
    )

  ;; Change switch entrance from one segment to another
  (:action change-switch-entrance
    :parameters (?sw - switch ?s-old ?s-new - segment)
    :precondition (and (switch-entrance ?sw ?s-old)
                       (SWITCH-HAS-ENTRANCE ?sw ?s-new)
                       (switch-available ?sw)
		    )
    :effect (and (not (switch-entrance ?sw ?s-old))
		 (switch-entrance ?sw ?s-new)
		 )
    )

  ;; Change switch exit from one segment to another
  (:action change-switch-exit
    :parameters (?sw - switch ?s-old ?s-new - segment)
    :precondition (and (switch-exit ?sw ?s-old)
                       (SWITCH-HAS-EXIT ?sw ?s-new)
                       (switch-available ?sw)
		    )
    :effect (and (not (switch-exit ?sw ?s-old))
		 (switch-exit ?sw ?s-new)
		 )
    )
  )